import logging
from typing import Any, Dict, List

import pytest
from packaging.version import Version
from packaging.version import parse as parse_version

from great_expectations.compatibility import pyspark
from great_expectations.datasource.fluent import SparkDatasource

logger = logging.getLogger(__name__)

try:
    from great_expectations.execution_engine import SparkDFExecutionEngine
except ImportError:
    SparkDFExecutionEngine = None
    # TODO: review logging more detail here
    logger.debug(
        "Unable to load pyspark; install optional spark dependency if you will be working with Spark dataframes."  # noqa: E501 # FIXME CoP
    )

# module level markers
pytestmark = pytest.mark.spark


def test_current_pyspark_version_installed(spark_session):
    pyspark_version: Version = parse_version(pyspark.pyspark.__version__)
    # Spark versions less than 3.0 are not supported.
    assert pyspark_version.major >= 3, "Spark versions less than 3.0 are not supported."


def test_spark_config_datasource(spark_session_v012):
    name = "great_expectations-ds-config"
    spark_config = {
        "spark.app.name": name,
        "spark.sql.catalogImplementation": "hive",
        "spark.executor.memory": "768m",
    }
    spark_datasource = SparkDatasource(
        name="my spark datasource",
        spark_config=spark_config,
    )
    # a warning is raised because passing unmodifiable config options results in restarting spark context  # noqa: E501 # FIXME CoP
    with pytest.warns(RuntimeWarning):
        execution_engine: SparkDFExecutionEngine = spark_datasource.get_execution_engine()
    spark_session: pyspark.SparkSession = execution_engine.spark
    sc_stopped: bool = spark_session.sparkContext._jsc.sc().isStopped()
    assert not sc_stopped

    # Test that our values were set
    conf: List[tuple] = spark_session.sparkContext.getConf().getAll()
    assert ("spark.app.name", name) in conf
    assert ("spark.sql.catalogImplementation", "hive") in conf
    assert ("spark.executor.memory", "768m") in conf
    spark_session.sparkContext.stop()


def test_spark_config_execution_engine_block_config(spark_session):
    new_spark_config: Dict[str, Any] = {
        "spark.app.name": "great_expectations-ee-config",
        "spark.sql.catalogImplementation": "hive",
        "spark.executor.memory": "512m",
    }
    with pytest.warns(RuntimeWarning):
        execution_engine = SparkDFExecutionEngine(spark_config=new_spark_config)
    new_spark_session: pyspark.SparkSession = execution_engine.spark

    # noinspection PyProtectedMember
    sc_stopped: bool = new_spark_session.sparkContext._jsc.sc().isStopped()

    assert not sc_stopped

    current_spark_config: List[tuple] = execution_engine.spark.sparkContext.getConf().getAll()
    assert ("spark.sql.catalogImplementation", "hive") in current_spark_config
    assert (
        "spark.app.name",
        "great_expectations-ee-config",
    ) in current_spark_config
    assert ("spark.executor.memory", "512m") in current_spark_config
    # spark context config values cannot be changed by the builder no matter what
    assert current_spark_config != new_spark_config
    new_spark_session.sparkContext.stop()



import pandas as pd

from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.column.distinct_values_count import (
    ColumnDistinctValuesCount,
    ColumnDistinctValuesCountResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import ALL_DATA_SOURCES

COLUMN_NAME = "whatevs"
DATA_FRAME = pd.DataFrame(
    {
        COLUMN_NAME: ["a", "b", "c", "c", "c", "c", "c", "c", "c", "c", None, None],
    },
)


class TestColumnDistinctValuesCount:
    @parameterize_batch_for_data_sources(
        data_source_configs=ALL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_distinct_values_count(self, batch_for_datasource: Batch) -> None:
        metric = ColumnDistinctValuesCount(column=COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnDistinctValuesCountResult)
        assert metric_result.value == 3



import pathlib

import pytest

from great_expectations.data_context.data_context.ephemeral_data_context import (
    EphemeralDataContext,
)
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from tests.test_utils import working_directory


@pytest.mark.filesystem
def test_convert_to_file_context(
    tmp_path: pathlib.Path,
    ephemeral_context_with_defaults: EphemeralDataContext,
):
    context = ephemeral_context_with_defaults

    datasource_name = "my_pandas"
    context.data_sources.add_pandas(datasource_name)

    tmp_path.mkdir(exist_ok=True)
    with working_directory(str(tmp_path)):
        migrated_context = context.convert_to_file_context()

    assert isinstance(migrated_context, FileDataContext)
    assert (
        len(migrated_context.data_sources.all()) == 1
        and datasource_name in migrated_context.data_sources.all()
    )



import sys
from pathlib import Path

import pytest

# Add the project root to the Python path so we can import from setup.py
project_root = Path(__file__).parent.parent
sys.path.insert(0, str(project_root))

from setup import parse_requirements


class TestParseRequirements:
    @pytest.mark.parametrize(
        "content,expected",
        [
            pytest.param(
                """numpy>=1.20.0
pandas>=1.3.0
requests>=2.25.0""",
                ["numpy>=1.20.0", "pandas>=1.3.0", "requests>=2.25.0"],
                id="simple_requirements",
            ),
            pytest.param(
                """# This is a comment
numpy>=1.20.0
# Another comment
pandas>=1.3.0
requests>=2.25.0""",
                ["numpy>=1.20.0", "pandas>=1.3.0", "requests>=2.25.0"],
                id="full_line_comments",
            ),
            pytest.param(
                """numpy>=1.20.0  # Scientific computing
pandas>=1.3.0  # Data manipulation
requests>=2.25.0  # HTTP library""",
                ["numpy>=1.20.0", "pandas>=1.3.0", "requests>=2.25.0"],
                id="inline_comments",
            ),
            pytest.param(
                """numpy>=1.20.0

pandas>=1.3.0


requests>=2.25.0""",
                ["numpy>=1.20.0", "pandas>=1.3.0", "requests>=2.25.0"],
                id="empty_lines",
            ),
            pytest.param(
                """# Main dependencies
numpy>=1.20.0  # Scientific computing

# Data processing
pandas>=1.3.0

# HTTP requests
requests>=2.25.0  # For API calls

# Optional dependencies
# scipy>=1.7.0""",
                ["numpy>=1.20.0", "pandas>=1.3.0", "requests>=2.25.0"],
                id="mixed_content",
            ),
            pytest.param(
                """numpy>=1.20.0,<2.0
pandas>=1.3.0,!=1.4.0
requests~=2.25.0
scipy==1.7.3
matplotlib>3.0,<=3.5.2""",
                [
                    "numpy>=1.20.0,<2.0",
                    "pandas>=1.3.0,!=1.4.0",
                    "requests~=2.25.0",
                    "scipy==1.7.3",
                    "matplotlib>3.0,<=3.5.2",
                ],
                id="complex_versions",
            ),
            pytest.param(
                """requests[security]>=2.25.0
sqlalchemy[postgresql,mysql]>=1.4.0
pytest[testing]>=6.0.0  # Testing framework""",
                [
                    "requests[security]>=2.25.0",
                    "sqlalchemy[postgresql,mysql]>=1.4.0",
                    "pytest[testing]>=6.0.0",
                ],
                id="requirements_with_extras",
            ),
            pytest.param("", [], id="empty_file"),
            pytest.param(
                """# This file contains only comments
# No actual requirements
# Another comment line""",
                [],
                id="only_comments",
            ),
            pytest.param(
                """  numpy>=1.20.0
    pandas>=1.3.0
requests>=2.25.0  # comment with spaces  """,
                ["numpy>=1.20.0", "pandas>=1.3.0", "requests>=2.25.0"],
                id="whitespace_handling",
            ),
        ],
    )
    @pytest.mark.unit
    def test_parse_requirements(self, content, expected, tmp_path):
        """Test parsing requirements files with various content formats."""
        # Create temporary requirements file
        requirements_file = tmp_path / "requirements.txt"
        requirements_file.write_text(content)

        result = parse_requirements(requirements_file)
        assert result == expected

    @pytest.mark.unit
    def test_parse_requirements_file_not_found(self):
        """Test that FileNotFoundError is raised for non-existent files."""
        with pytest.raises(FileNotFoundError):
            parse_requirements(Path("/nonexistent/path/requirements.txt"))



import re
from datetime import datetime
from pprint import pprint
from typing import Any, Callable, Dict, Optional, Tuple, Union

import pytest

from great_expectations.core import ExpectationValidationResult
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.expectations.registry import get_renderer_impl
from great_expectations.render import RenderedAtomicContent
from great_expectations.render.renderer.observed_value_renderer import ObservedValueRenderState


@pytest.fixture
def expectation_configuration_kwargs():
    # These below fields are defaults; specific tests will overwrite as deemed necessary
    return {
        "id": "abcdefgh-ijkl-mnop-qrst-uvwxyz123456",
        "type": "",
        "kwargs": {},
        "meta": {},
    }


@pytest.fixture
def get_prescriptive_rendered_content(
    expectation_configuration_kwargs: Dict[str, Union[str, dict]],
) -> Callable:
    def _get_prescriptive_rendered_content(
        update_dict: Dict[str, Union[str, dict]],
    ) -> RenderedAtomicContent:
        # Overwrite any fields passed in from test and instantiate ExpectationConfiguration
        expectation_configuration_kwargs.update(update_dict)
        config = ExpectationConfiguration(**expectation_configuration_kwargs)
        expectation_type = expectation_configuration_kwargs["type"]

        # Programatically determine the renderer implementations
        renderer_impl = get_renderer_impl(
            object_name=expectation_type,
            renderer_type="atomic.prescriptive.summary",
        )[1]

        # Determine RenderedAtomicContent output
        source_obj = {"configuration": config}
        res = renderer_impl(**source_obj)
        return res

    return _get_prescriptive_rendered_content


@pytest.fixture
def evr_kwargs(expectation_configuration_kwargs):
    # These below fields are defaults; specific tests will overwrite as deemed necessary
    return {
        "expectation_config": ExpectationConfiguration(**expectation_configuration_kwargs),
        "result": {},
    }


@pytest.fixture
def get_diagnostic_rendered_content(
    evr_kwargs: Dict[str, Union[dict, ExpectationConfiguration]],
) -> Callable:
    def _get_diagnostic_rendered_content(
        update_dict: Dict[str, Union[dict, ExpectationConfiguration]],
    ) -> RenderedAtomicContent:
        # Overwrite any fields passed in from test and instantiate ExpectationValidationResult
        evr_kwargs.update(update_dict)
        evr = ExpectationValidationResult(**evr_kwargs)
        expectation_config = evr_kwargs["expectation_config"]
        expectation_type = expectation_config["type"]

        # Programatically determine the renderer implementations
        renderer_impl = get_renderer_impl(
            object_name=expectation_type,
            renderer_type="atomic.diagnostic.observed_value",
        )[1]

        # Determine RenderedAtomicContent output
        source_obj = {"result": evr}
        res = renderer_impl(**source_obj)
        return res

    return _get_diagnostic_rendered_content


# "atomic.prescriptive.summary" tests


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_bootstrapped_ks_test_p_value_to_be_greater_than(
    get_prescriptive_rendered_content,
):
    # Expectation is a stub; open to implement test once renderer method is available
    pass


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_chisquare_test_p_value_to_be_greater_than(
    get_prescriptive_rendered_content,
):
    # Expectation is a stub; open to implement test once renderer method is available
    pass


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_distinct_values_to_be_in_set(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_distinct_values_to_be_in_set",
        "kwargs": {
            "column": "my_column",
            "value_set": [1, 2, 3],
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "v__0": {"schema": {"type": "number"}, "value": 1},
                "v__1": {"schema": {"type": "number"}, "value": 2},
                "v__2": {"schema": {"type": "number"}, "value": 3},
                "value_set": {"schema": {"type": "array"}, "value": [1, 2, 3]},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column distinct values must belong to this set: $v__0 $v__1 $v__2.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_distinct_values_to_contain_set(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_distinct_values_to_contain_set",
        "kwargs": {
            "column": "my_column",
            "value_set": ["a", "b", "c"],
            "parse_strings_as_datetimes": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "v__0": {"schema": {"type": "string"}, "value": "a"},
                "v__1": {"schema": {"type": "string"}, "value": "b"},
                "v__2": {"schema": {"type": "string"}, "value": "c"},
                "value_set": {"schema": {"type": "array"}, "value": ["a", "b", "c"]},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column distinct values must contain this set: $v__0 $v__1 $v__2.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_distinct_values_to_equal_set(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_distinct_values_to_equal_set",
        "kwargs": {
            "column": "my_column",
            "value_set": ["a", "b", "c"],
            "parse_strings_as_datetimes": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "v__0": {"schema": {"type": "string"}, "value": "a"},
                "v__1": {"schema": {"type": "string"}, "value": "b"},
                "v__2": {"schema": {"type": "string"}, "value": "c"},
                "value_set": {"schema": {"type": "array"}, "value": ["a", "b", "c"]},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column distinct values must match this set: $v__0 $v__1 $v__2.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_kl_divergence_to_be_less_than(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_kl_divergence_to_be_less_than",
        "kwargs": {
            "column": "min_event_time",
            "partition_object": {
                "bins": [0, 5, 10, 30, 50],
                "weights": [0.2, 0.3, 0.1, 0.4],
            },
            "threshold": 0.1,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)

    # replace version of vega-lite in res to match snapshot test
    res["value"]["graph"]["$schema"] = re.sub(
        r"v\d*\.\d*\.\d*", "v4.8.1", res["value"]["graph"]["$schema"]
    )

    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "graph": {
                "$schema": "https://vega.github.io/schema/vega-lite/v4.8.1.json",
                "autosize": "fit",
                "config": {"view": {"continuousHeight": 300, "continuousWidth": 400}},
                "data": {"name": "data-1cb20570b53cc3e67cb4883fd45e64cb"},
                "datasets": {
                    "data-1cb20570b53cc3e67cb4883fd45e64cb": [
                        {"bin_max": 5, "bin_min": 0, "fraction": 0.2},
                        {"bin_max": 10, "bin_min": 5, "fraction": 0.3},
                        {"bin_max": 30, "bin_min": 10, "fraction": 0.1},
                        {"bin_max": 50, "bin_min": 30, "fraction": 0.4},
                    ]
                },
                "encoding": {
                    "tooltip": [
                        {"field": "bin_min", "type": "quantitative"},
                        {"field": "bin_max", "type": "quantitative"},
                        {"field": "fraction", "type": "quantitative"},
                    ],
                    "x": {"field": "bin_min", "type": "ordinal"},
                    "x2": {"field": "bin_max"},
                    "y": {"field": "fraction", "type": "quantitative"},
                },
                "height": 400,
                "mark": "bar",
                "width": 250,
            },
            "header": {
                "schema": {"type": "StringValueType"},
                "value": {
                    "params": {
                        "column": {"schema": {"type": "string"}, "value": "min_event_time"},
                        "threshold": {"schema": {"type": "number"}, "value": 0.1},
                    },
                    "template": "$column Kullback-Leibler (KL) divergence with respect to the following distribution must be lower than $threshold.",  # noqa: E501 # FIXME CoP
                },
            },
            "schema": {"type": "GraphType"},
        },
        "value_type": "GraphType",
    }


@pytest.mark.unit
def test_atomic_diagnostic_observed_value_expect_column_kl_divergence_to_be_less_than(
    get_diagnostic_rendered_content,
):
    # Please note that the vast majority of Expectations are calling `Expectation._atomic_diagnostic_observed_value()`  # noqa: E501 # FIXME CoP
    # As such, the specific expectation_type used here is irrelevant and is simply used to trigger the parent class.  # noqa: E501 # FIXME CoP
    expectation_config = {
        "type": "expect_column_kl_divergence_to_be_less_than",
        "kwargs": {
            "column": "min_event_time",
            "partition_object": {
                "bins": [0, 5, 10, 30, 50],
                "weights": [0.2, 0.3, 0.1, 0.4],
            },
            "threshold": 0.1,
        },
        "meta": {},
        "id": "4b53c4d5-90ba-467a-b7a7-379640bbd729",
    }
    update_dict = {
        "expectation_config": ExpectationConfiguration(**expectation_config),
        "result": {
            "observed_value": 0.0,
            "details": {
                "observed_partition": {
                    "values": [1, 2, 4],
                    "weights": [0.3754, 0.615, 0.0096],
                },
                "expected_partition": {
                    "values": [1, 2, 4],
                    "weights": [0.3754, 0.615, 0.0096],
                },
            },
        },
    }
    rendered_content = get_diagnostic_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)

    # replace version of vega-lite in res to match snapshot test
    res["value"]["graph"]["$schema"] = re.sub(
        r"v\d*\.\d*\.\d*", "v4.8.1", res["value"]["graph"]["$schema"]
    )
    assert res == {
        "name": "atomic.diagnostic.observed_value",
        "value": {
            "graph": {
                "$schema": "https://vega.github.io/schema/vega-lite/v4.8.1.json",
                "autosize": "fit",
                "config": {"view": {"continuousHeight": 300, "continuousWidth": 400}},
                "data": {"name": "data-c49f2d3d7bdab36f9ec81f5a314a430c"},
                "datasets": {
                    "data-c49f2d3d7bdab36f9ec81f5a314a430c": [
                        {"fraction": 0.3754, "values": 1},
                        {"fraction": 0.615, "values": 2},
                        {"fraction": 0.0096, "values": 4},
                    ]
                },
                "encoding": {
                    "tooltip": [
                        {"field": "values", "type": "quantitative"},
                        {"field": "fraction", "type": "quantitative"},
                    ],
                    "x": {"field": "values", "type": "nominal"},
                    "y": {"field": "fraction", "type": "quantitative"},
                },
                "height": 400,
                "mark": "bar",
                "width": 250,
            },
            "header": {
                "schema": {"type": "StringValueType"},
                "value": {
                    "params": {
                        "observed_value": {
                            "schema": {"type": "string"},
                            "value": "None (-infinity, infinity, or NaN)",
                        }
                    },
                    "template": "KL Divergence: $observed_value",
                },
            },
            "schema": {"type": "GraphType"},
        },
        "value_type": "GraphType",
    }


@pytest.mark.unit
def test_atomic_diagnostic_observed_value_with_boolean_column_expect_column_kl_divergence_to_be_less_than(  # noqa: E501 # FIXME CoP
    get_diagnostic_rendered_content,
):
    # Please note that the vast majority of Expectations are calling `Expectation._atomic_diagnostic_observed_value()`  # noqa: E501 # FIXME CoP
    # As such, the specific expectation_type used here is irrelevant and is simply used to trigger the parent class.  # noqa: E501 # FIXME CoP
    expectation_config = {
        "type": "expect_column_kl_divergence_to_be_less_than",
        "kwargs": {
            "column": "boolean_event",
            "partition_object": {
                "bins": [True, False],
                "weights": [0.5, 0.5],
            },
            "threshold": 0.1,
        },
        "meta": {},
        "id": "4b53c4d5-90ba-467a-b7a7-379640bbd729",
    }
    update_dict = {
        "expectation_config": ExpectationConfiguration(**expectation_config),
        "result": {
            "observed_value": 0.0,
            "details": {
                "observed_partition": {
                    "values": [True, False],
                    "weights": [0.5, 0.5],
                },
                "expected_partition": {
                    "values": [True, False],
                    "weights": [0.5, 0.5],
                },
            },
        },
    }
    rendered_content = get_diagnostic_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)

    # replace version of vega-lite in res to match snapshot test
    res["value"]["graph"]["$schema"] = re.sub(
        r"v\d*\.\d*\.\d*", "v4.8.1", res["value"]["graph"]["$schema"]
    )
    assert res == {
        "name": "atomic.diagnostic.observed_value",
        "value": {
            "graph": {
                "$schema": "https://vega.github.io/schema/vega-lite/v4.8.1.json",
                "autosize": "fit",
                "config": {"view": {"continuousHeight": 300, "continuousWidth": 400}},
                "data": {"name": "data-d8f1a1ab1f79e142d9ca399157673554"},
                "datasets": {
                    "data-d8f1a1ab1f79e142d9ca399157673554": [
                        {"fraction": 0.5, "values": "True"},
                        {"fraction": 0.5, "values": "False"},
                    ]
                },
                "encoding": {
                    "tooltip": [
                        {"field": "values", "type": "nominal"},
                        {"field": "fraction", "type": "quantitative"},
                    ],
                    "x": {"field": "values", "type": "nominal"},
                    "y": {"field": "fraction", "type": "quantitative"},
                },
                "height": 400,
                "mark": "bar",
                "width": 250,
            },
            "header": {
                "schema": {"type": "StringValueType"},
                "value": {
                    "params": {
                        "observed_value": {
                            "schema": {"type": "string"},
                            "value": "None (-infinity, infinity, or NaN)",
                        }
                    },
                    "template": "KL Divergence: $observed_value",
                },
            },
            "schema": {"type": "GraphType"},
        },
        "value_type": "GraphType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_max_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_max_to_be_between",
        "kwargs": {
            "column": "my_column",
            "min_value": 1,
            "max_value": 5,
            "parse_strings_as_datetimes": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "max_value": {"schema": {"type": "number"}, "value": 5},
                "min_value": {"schema": {"type": "number"}, "value": 1},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column maximum value must be greater than or equal to $min_value and less than or equal to $max_value.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_mean_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_mean_to_be_between",
        "kwargs": {
            "column": "my_column",
            "min_value": 3,
            "max_value": 7,
            "parse_strings_as_datetimes": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "max_value": {"schema": {"type": "number"}, "value": 7},
                "min_value": {"schema": {"type": "number"}, "value": 3},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column mean must be greater than or equal to $min_value and less than or equal to $max_value.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_median_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_median_to_be_between",
        "kwargs": {
            "column": "my_column",
            "min_value": 5,
            "max_value": 10,
            "parse_strings_as_datetimes": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "max_value": {"schema": {"type": "number"}, "value": 10},
                "min_value": {"schema": {"type": "number"}, "value": 5},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column median must be greater than or equal to $min_value and less than or equal to $max_value.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_min_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_min_to_be_between",
        "kwargs": {
            "column": "my_column",
            "min_value": 1,
            "max_value": 5,
            "parse_strings_as_datetimes": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "max_value": {"schema": {"type": "number"}, "value": 5},
                "min_value": {"schema": {"type": "number"}, "value": 1},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column minimum value must be greater than or equal to $min_value and less than or equal to $max_value.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_most_common_value_to_be_in_set(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_most_common_value_to_be_in_set",
        "kwargs": {
            "column": "my_column",
            "value_set": [1, 2, 3],
            "ties_okay": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "ties_okay": {"schema": {"type": "boolean"}, "value": True},
                "v__0": {"schema": {"type": "number"}, "value": 1},
                "v__1": {"schema": {"type": "number"}, "value": 2},
                "v__2": {"schema": {"type": "number"}, "value": 3},
                "value_set": {"schema": {"type": "array"}, "value": [1, 2, 3]},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column most common value must belong to this set: $v__0 $v__1 $v__2. Values outside this set that are as common (but not more common) are allowed.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
@pytest.mark.xfail(
    strict=False,
    reason="ExpectColumnPairCramersPhiValueToBeLessThan is not fully implemented",
)
def test_atomic_prescriptive_summary_expect_column_pair_cramers_phi_value_to_be_less_than(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_pair_cramers_phi_value_to_be_less_than",
        "kwargs": {
            "column_A": "foo",
            "column_B": "bar",
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {}


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_pair_values_a_to_be_greater_than_b(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_pair_values_a_to_be_greater_than_b",
        "kwargs": {
            "column_A": "foo",
            "column_B": "bar",
            "parse_strings_as_datetimes": True,
            "mostly": 0.8,
            "ignore_row_if": "baz",
            "or_equal": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column_A": {"schema": {"type": "string"}, "value": "foo"},
                "column_B": {"schema": {"type": "string"}, "value": "bar"},
                "ignore_row_if": {"schema": {"type": "string"}, "value": "baz"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "or_equal": {"schema": {"type": "boolean"}, "value": True},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Values in $column_A must be greater than or equal to those in $column_B, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_pair_values_to_be_equal(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_pair_values_to_be_equal",
        "kwargs": {
            "column_A": "foo",
            "column_B": "bar",
            "mostly": 0.8,
            "ignore_row_if": "baz",
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column_A": {"schema": {"type": "string"}, "value": "foo"},
                "column_B": {"schema": {"type": "string"}, "value": "bar"},
                "ignore_row_if": {"schema": {"type": "string"}, "value": "baz"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Values in $column_A and $column_B must be equal, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_pair_values_to_be_in_set(
    get_prescriptive_rendered_content,
):
    # Expectation is a stub; open to implement test once renderer method is available
    pass


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_parameterized_distribution_ks_test_p_value_to_be_greater_than(  # noqa: E501 # FIXME CoP
    get_prescriptive_rendered_content,
):
    # Expectation is a stub; open to implement test once renderer method is available
    pass


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_proportion_of_unique_values_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_proportion_of_unique_values_to_be_between",
        "kwargs": {
            "column": "my_column",
            "min_value": 10,
            "max_value": 20,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "max_value": {"schema": {"type": "number"}, "value": 20},
                "min_value": {"schema": {"type": "number"}, "value": 10},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column proportion of unique values must be greater than or equal to $min_value and less than or equal to $max_value.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_quantile_values_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_quantile_values_to_be_between",
        "kwargs": {
            "column": "Unnamed: 0",
            "quantile_ranges": {
                "quantiles": [0.05, 0.25, 0.5, 0.75, 0.95],
                "value_ranges": [
                    [66, 68],
                    [328, 330],
                    [656, 658],
                    [984, 986],
                    [1246, 1248],
                ],
            },
            "allow_relative_error": False,
        },
        "meta": {},
        "id": "cd6b4f19-8167-4984-b495-54bffcb070da",
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "header": {
                "schema": {"type": "StringValueType"},
                "value": {
                    "params": {"column": {"schema": {"type": "string"}, "value": "Unnamed: 0"}},
                    "template": "$column quantiles must be within the following value ranges.",
                },
            },
            "header_row": [
                {"schema": {"type": "string"}, "value": "Quantile"},
                {"schema": {"type": "string"}, "value": "Min Value"},
                {"schema": {"type": "string"}, "value": "Max Value"},
            ],
            "schema": {"type": "TableType"},
            "table": [
                [
                    {"schema": {"type": "string"}, "value": "0.05"},
                    {"schema": {"type": "number"}, "value": 66},
                    {"schema": {"type": "number"}, "value": 68},
                ],
                [
                    {"schema": {"type": "string"}, "value": "Q1"},
                    {"schema": {"type": "number"}, "value": 328},
                    {"schema": {"type": "number"}, "value": 330},
                ],
                [
                    {"schema": {"type": "string"}, "value": "Median"},
                    {"schema": {"type": "number"}, "value": 656},
                    {"schema": {"type": "number"}, "value": 658},
                ],
                [
                    {"schema": {"type": "string"}, "value": "Q3"},
                    {"schema": {"type": "number"}, "value": 984},
                    {"schema": {"type": "number"}, "value": 986},
                ],
                [
                    {"schema": {"type": "string"}, "value": "0.95"},
                    {"schema": {"type": "number"}, "value": 1246},
                    {"schema": {"type": "number"}, "value": 1248},
                ],
            ],
        },
        "value_type": "TableType",
    }


@pytest.mark.unit
def test_atomic_diagnostic_observed_value_expect_column_quantile_values_to_be_between(
    get_diagnostic_rendered_content,
):
    # Please note that the vast majority of Expectations are calling `Expectation._atomic_diagnostic_observed_value()`  # noqa: E501 # FIXME CoP
    # As such, the specific expectation_type used here is irrelevant and is simply used to trigger the parent class.  # noqa: E501 # FIXME CoP
    expectation_config = {
        "type": "expect_column_quantile_values_to_be_between",
        "kwargs": {
            "column": "Unnamed: 0",
            "quantile_ranges": {
                "quantiles": [0.05, 0.25, 0.5, 0.75, 0.95],
                "value_ranges": [
                    [66, 68],
                    [328, 330],
                    [656, 658],
                    [984, 986],
                    [1246, 1248],
                ],
            },
            "allow_relative_error": False,
        },
        "meta": {},
        "id": "cd6b4f19-8167-4984-b495-54bffcb070da",
    }
    update_dict = {
        "expectation_config": ExpectationConfiguration(**expectation_config),
        "result": {
            "observed_value": {
                "quantiles": [0.05, 0.25, 0.5, 0.75, 0.95],
                "values": [67, 329, 657, 985, 1247],
            },
            "element_count": 1313,
            "missing_count": None,
            "missing_percent": None,
            "details": {"success_details": [True, True, True, True, True]},
        },
    }
    rendered_content = get_diagnostic_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.diagnostic.observed_value",
        "value": {
            "header_row": [
                {"schema": {"type": "string"}, "value": "Quantile"},
                {"schema": {"type": "string"}, "value": "Value"},
            ],
            "schema": {"type": "TableType"},
            "table": [
                [
                    {"schema": {"type": "string"}, "value": "0.05"},
                    {"schema": {"type": "number"}, "value": 67},
                ],
                [
                    {"schema": {"type": "string"}, "value": "Q1"},
                    {"schema": {"type": "number"}, "value": 329},
                ],
                [
                    {"schema": {"type": "string"}, "value": "Median"},
                    {"schema": {"type": "number"}, "value": 657},
                ],
                [
                    {"schema": {"type": "string"}, "value": "Q3"},
                    {"schema": {"type": "number"}, "value": 985},
                ],
                [
                    {"schema": {"type": "string"}, "value": "0.95"},
                    {"schema": {"type": "number"}, "value": 1247},
                ],
            ],
        },
        "value_type": "TableType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_stdev_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_stdev_to_be_between",
        "kwargs": {
            "column": "my_column",
            "min_value": 10,
            "max_value": 20,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "max_value": {"schema": {"type": "number"}, "value": 20},
                "min_value": {"schema": {"type": "number"}, "value": 10},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column standard deviation must be greater than or equal to $min_value and less than or equal to $max_value.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_sum_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_sum_to_be_between",
        "kwargs": {
            "column": "my_column",
            "min_value": 10,
            "max_value": 20,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "max_value": {"schema": {"type": "number"}, "value": 20},
                "min_value": {"schema": {"type": "number"}, "value": 10},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column sum must be greater than or equal to $min_value and less than or equal to $max_value.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_to_exist(get_prescriptive_rendered_content):
    update_dict = {
        "type": "expect_column_to_exist",
        "kwargs": {
            "column": "my_column",
            "column_index": 5,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "column_index": {"schema": {"type": "number"}, "value": 5},
                "column_indexth": {"schema": {"type": "string"}, "value": "5th"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column must be the $column_indexth field.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_unique_value_count_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_unique_value_count_to_be_between",
        "kwargs": {
            "column": "my_column",
            "mostly": 0.8,
            "min_value": 10,
            "max_value": 20,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "max_value": {"schema": {"type": "number"}, "value": 20},
                "min_value": {"schema": {"type": "number"}, "value": 10},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column must have greater than or equal to $min_value and less than or equal to $max_value unique values.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_value_lengths_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_value_lengths_to_be_between",
        "kwargs": {
            "column": "my_column",
            "mostly": 0.8,
            "min_value": 10,
            "max_value": 20,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "max_value": {"schema": {"type": "number"}, "value": 20},
                "min_value": {"schema": {"type": "number"}, "value": 10},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must be greater than or equal to $min_value and less than or equal to $max_value characters long, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_value_lengths_to_equal(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_value_lengths_to_equal",
        "kwargs": {
            "column": "my_column",
            "value": 100,
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "value": {"schema": {"type": "number"}, "value": 100},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must be $value characters long, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_value_z_scores_to_be_less_than(
    get_prescriptive_rendered_content,
):
    # Expectation is a stub; open to implement test once renderer method is available
    pass


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_be_between",
        "kwargs": {
            "column": "my_column",
            "mostly": 0.8,
            "min_value": 1,
            "max_value": 5,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "max_value": {"schema": {"type": "number"}, "value": 5},
                "min_value": {"schema": {"type": "number"}, "value": 1},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must be greater than or equal to $min_value and less than or equal to $max_value, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_be_dateutil_parseable(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_be_dateutil_parseable",
        "kwargs": {
            "column": "my_column",
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must be parseable by dateutil, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_be_decreasing(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_be_decreasing",
        "kwargs": {
            "column": "my_column",
            "mostly": 0.8,
            "parse_strings_as_datetimes": True,
            "strictly": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "strictly": {"schema": {"type": "boolean"}, "value": True},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must be strictly less than previous values, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_be_in_set(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_be_in_set",
        "kwargs": {
            "column": "my_column",
            "value_set": [1, 2, 3, 4],
            "mostly": 0.8,
            "parse_strings_as_datetimes": True,
            "strictly": 50,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "v__0": {"schema": {"type": "number"}, "value": 1},
                "v__1": {"schema": {"type": "number"}, "value": 2},
                "v__2": {"schema": {"type": "number"}, "value": 3},
                "v__3": {"schema": {"type": "number"}, "value": 4},
                "value_set": {"schema": {"type": "array"}, "value": [1, 2, 3, 4]},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must belong to this set: $v__0 $v__1 $v__2 $v__3, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_be_in_type_list(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_be_in_type_list",
        "kwargs": {
            "column": "my_column",
            "type_list": ["type_a", "type_b", "type_c"],
            "value_set": [1, 2, 3, 4],
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "type_list": {
                    "schema": {"type": "array"},
                    "value": ["type_a", "type_b", "type_c"],
                },
                "v__0": {"schema": {"type": "string"}, "value": "type_a"},
                "v__1": {"schema": {"type": "string"}, "value": "type_b"},
                "v__2": {"schema": {"type": "string"}, "value": "type_c"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column value types must belong to this set: $v__0 $v__1 $v__2, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_be_increasing(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_be_increasing",
        "kwargs": {
            "column": "my_column",
            "strictly": True,
            "mostly": 0.8,
            "parse_strings_as_datetimes": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "strictly": {"schema": {"type": "boolean"}, "value": True},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must be strictly greater than previous values, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_be_json_parseable(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_be_json_parseable",
        "kwargs": {
            "column": "my_column",
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must be parseable as JSON, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_be_null(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_be_null",
        "kwargs": {
            "column": "my_column",
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must be null, at least $mostly_pct % of the time.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_be_null_with_mostly_equals_1(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_be_null",
        "kwargs": {
            "column": "my_column",
            "mostly": 1.0,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 1.0},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must be null.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_be_of_type(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_be_of_type",
        "kwargs": {
            "column": "my_column",
            "type_": "my_type",
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "type_": {"schema": {"type": "string"}, "value": "my_type"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must be of type $type_, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_be_unique(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_be_unique",
        "kwargs": {
            "column": "my_column",
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must be unique, at least $mostly_pct % of the time.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_match_json_schema(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_match_json_schema",
        "kwargs": {
            "column": "my_column",
            "mostly": 0.8,
            "json_schema": {"foo": "bar"},
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "formatted_json": {
                    "schema": {"type": "string"},
                    "value": """<pre>{\n    "foo": "bar"\n}</pre>""",
                },
                "json_schema": {"schema": {"type": "object"}, "value": {"foo": "bar"}},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must match the following JSON Schema, at least $mostly_pct % of the time: $formatted_json",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_match_like_pattern(
    get_prescriptive_rendered_content,
):
    # Expectation is a stub; open to implement test once renderer method is available
    pass


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_match_like_pattern_list(
    get_prescriptive_rendered_content,
):
    # Expectation is a stub; open to implement test once renderer method is available
    pass


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_match_regex(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_match_regex",
        "kwargs": {
            "column": "my_column",
            "mostly": 0.8,
            "regex": "^superconductive$",
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "regex": {"schema": {"type": "string"}, "value": "^superconductive$"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must match this regular expression: $regex, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_match_regex_list(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_match_regex_list",
        "kwargs": {
            "column": "my_column",
            "regex_list": ["^superconductive$", "ge|great_expectations"],
            "mostly": 0.8,
            "match_on": "all",
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "match_on": {"schema": {"type": "string"}, "value": "all"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "regex_list": {
                    "schema": {"type": "array"},
                    "value": ["^superconductive$", "ge|great_expectations"],
                },
                "v__0": {"schema": {"type": "string"}, "value": "^superconductive$"},
                "v__1": {"schema": {"type": "string"}, "value": "ge|great_expectations"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must match all of the following regular expressions: $v__0 $v__1, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_match_strftime_format(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_match_strftime_format",
        "kwargs": {
            "column": "my_column",
            "strftime_format": "%Y-%m",
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "strftime_format": {"schema": {"type": "string"}, "value": "%Y-%m"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must match the following strftime format: $strftime_format, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_not_be_in_set(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_not_be_in_set",
        "kwargs": {
            "column": "my_column",
            "value_set": [1, 2, 3],
            "mostly": 0.8,
            "parse_string_as_datetimes": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "v__0": {"schema": {"type": "number"}, "value": 1},
                "v__1": {"schema": {"type": "number"}, "value": 2},
                "v__2": {"schema": {"type": "number"}, "value": 3},
                "value_set": {"schema": {"type": "array"}, "value": [1, 2, 3]},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must not belong to this set: $v__0 $v__1 $v__2, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_not_be_null(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_not_be_null",
        "kwargs": {
            "column": "my_column",
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must not be null, at least $mostly_pct % of the time.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_not_match_like_pattern(
    get_prescriptive_rendered_content,
):
    # Expectation is a stub; open to implement test once renderer method is available
    pass


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_not_match_like_pattern_list(
    get_prescriptive_rendered_content,
):
    # Expectation is a stub; open to implement test once renderer method is available
    pass


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_not_match_regex(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_not_match_regex",
        "kwargs": {
            "column": "my_column",
            "regex": "^superconductive$",
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "regex": {"schema": {"type": "string"}, "value": "^superconductive$"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must not match this regular expression: $regex, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_column_values_to_not_match_regex_list(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_column_values_to_not_match_regex_list",
        "kwargs": {
            "column": "my_column",
            "regex_list": ["^a", "^b", "^c"],
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
                "regex_list": {"schema": {"type": "array"}, "value": ["^a", "^b", "^c"]},
                "v__0": {"schema": {"type": "string"}, "value": "^a"},
                "v__1": {"schema": {"type": "string"}, "value": "^b"},
                "v__2": {"schema": {"type": "string"}, "value": "^c"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$column values must not match any of the following regular expressions: $v__0 $v__1 $v__2, at least $mostly_pct % of the time.",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_compound_columns_to_be_unique(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_compound_columns_to_be_unique",
        "kwargs": {
            "column_list": ["my_first_col", "my_second_col", "my_third_col"],
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column_list": {
                    "schema": {"type": "array"},
                    "value": ["my_first_col", "my_second_col", "my_third_col"],
                },
                "column_list_0": {"schema": {"type": "string"}, "value": "my_first_col"},
                "column_list_1": {"schema": {"type": "string"}, "value": "my_second_col"},
                "column_list_2": {"schema": {"type": "string"}, "value": "my_third_col"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Values for given compound columns must be unique together, at least $mostly_pct % of the time: $column_list_0 $column_list_1 $column_list_2",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_multicolumn_sum_to_equal(
    get_prescriptive_rendered_content,
):
    # Expectation is a stub; open to implement test once renderer method is available
    pass


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_multicolumn_values_to_be_unique(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_multicolumn_values_to_be_unique",
        "kwargs": {
            "column_list": ["A", "B", "C"],
            "ignore_row_if": "foo",
            "mostly": 0.8,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column_list": {"schema": {"type": "array"}, "value": ["A", "B", "C"]},
                "column_list_0": {"schema": {"type": "string"}, "value": "A"},
                "column_list_1": {"schema": {"type": "string"}, "value": "B"},
                "column_list_2": {"schema": {"type": "string"}, "value": "C"},
                "ignore_row_if": {"schema": {"type": "string"}, "value": "foo"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Values must be unique across columns, at least $mostly_pct % of the time: $column_list_0 $column_list_1 $column_list_2",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_select_column_values_to_be_unique_within_record(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_select_column_values_to_be_unique_within_record",
        "kwargs": {
            "column_list": ["my_first_column", "my_second_column"],
            "mostly": 0.8,
            "ignore_row_if": "foo",
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column_list": {
                    "schema": {"type": "array"},
                    "value": ["my_first_column", "my_second_column"],
                },
                "column_list_0": {"schema": {"type": "string"}, "value": "my_first_column"},
                "column_list_1": {
                    "schema": {"type": "string"},
                    "value": "my_second_column",
                },
                "ignore_row_if": {"schema": {"type": "string"}, "value": "foo"},
                "mostly": {"schema": {"type": "number"}, "value": 0.8},
                "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Values must be unique across columns, at least $mostly_pct % of the time: $column_list_0 $column_list_1",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_table_column_count_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_table_column_count_to_be_between",
        "kwargs": {
            "min_value": 5,
            "max_value": None,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {"min_value": {"schema": {"type": "number"}, "value": 5}},
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Must have greater than or equal to $min_value columns.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_table_column_count_to_equal(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_table_column_count_to_equal",
        "kwargs": {
            "value": 10,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {"value": {"schema": {"type": "number"}, "value": 10}},
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Must have exactly $value columns.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_table_columns_to_match_ordered_list(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_table_columns_to_match_ordered_list",
        "kwargs": {"column_list": ["a", "b", "c"]},
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column_list": {"schema": {"type": "array"}, "value": ["a", "b", "c"]},
                "column_list_0": {"schema": {"type": "string"}, "value": "a"},
                "column_list_1": {"schema": {"type": "string"}, "value": "b"},
                "column_list_2": {"schema": {"type": "string"}, "value": "c"},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Must have these columns in this order: $column_list_0 $column_list_1 $column_list_2",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_table_columns_to_match_set(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_table_columns_to_match_set",
        "kwargs": {
            "column_set": ["a", "b", "c"],
            "exact_match": True,
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column_set": {"schema": {"type": "array"}, "value": ["a", "b", "c"]},
                "column_set_0": {"schema": {"type": "string"}, "value": "a"},
                "column_set_1": {"schema": {"type": "string"}, "value": "b"},
                "column_set_2": {"schema": {"type": "string"}, "value": "c"},
                "exact_match": {"schema": {"type": "boolean"}, "value": True},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Must have exactly these columns (in any order): $column_set_0 $column_set_1 $column_set_2",  # noqa: E501 # FIXME CoP
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_table_row_count_to_be_between(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_table_row_count_to_be_between",
        "kwargs": {"max_value": None, "min_value": 1},
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {"min_value": {"schema": {"type": "number"}, "value": 1}},
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Must have greater than or equal to $min_value rows.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_table_row_count_to_equal(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_table_row_count_to_equal",
        "kwargs": {"value": 10},
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {"value": {"schema": {"type": "number"}, "value": 10}},
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Must have exactly $value rows.",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_prescriptive_summary_expect_table_row_count_to_equal_other_table(
    get_prescriptive_rendered_content,
):
    update_dict = {
        "type": "expect_table_row_count_to_equal_other_table",
        "kwargs": {
            "other_table_name": {
                "schema": {"type": "string"},
                "value": "other_table_name",
            }
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "other_table_name": {
                    "schema": {"type": "string"},
                    "value": {"schema": {"type": "string"}, "value": "other_table_name"},
                }
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "Row count must equal the row count of table $other_table_name.",
        },
        "value_type": "StringValueType",
    }


# "atomic.diagnostic.observed_value" tests


@pytest.mark.unit
def test_atomic_diagnostic_observed_value_without_result(get_diagnostic_rendered_content):
    # Please note that the vast majority of Expectations are calling `Expectation._atomic_diagnostic_observed_value()`  # noqa: E501 # FIXME CoP
    # As such, the specific expectation_type used here is irrelevant and is simply used to trigger the parent class.  # noqa: E501 # FIXME CoP
    expectation_config = {
        "type": "expect_table_row_count_to_equal",
        "kwargs": {},
    }
    update_dict = {
        "expectation_config": ExpectationConfiguration(**expectation_config),
    }
    rendered_content = get_diagnostic_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.diagnostic.observed_value",
        "value": {
            "params": {"observed_value": {"schema": {"type": "string"}, "value": "--"}},
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$observed_value",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_diagnostic_observed_value_with_numeric_observed_value(
    get_diagnostic_rendered_content,
):
    # Please note that the vast majority of Expectations are calling `Expectation._atomic_diagnostic_observed_value()`  # noqa: E501 # FIXME CoP
    # As such, the specific expectation_type used here is irrelevant and is simply used to trigger the parent class.  # noqa: E501 # FIXME CoP
    expectation_config = {
        "type": "expect_table_row_count_to_equal",
        "kwargs": {},
    }
    update_dict = {
        "expectation_config": ExpectationConfiguration(**expectation_config),
        "result": {"observed_value": 1776},
    }
    rendered_content = get_diagnostic_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.diagnostic.observed_value",
        "value": {
            "params": {"observed_value": {"schema": {"type": "number"}, "value": 1776}},
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$observed_value",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_diagnostic_observed_value_with_str_observed_value(get_diagnostic_rendered_content):
    # Please note that the vast majority of Expectations are calling `Expectation._atomic_diagnostic_observed_value()`  # noqa: E501 # FIXME CoP
    # As such, the specific expectation_type used here is irrelevant and is simply used to trigger the parent class.  # noqa: E501 # FIXME CoP
    expectation_config = {
        "type": "expect_table_row_count_to_equal",
        "kwargs": {},
    }
    update_dict = {
        "expectation_config": ExpectationConfiguration(**expectation_config),
        "result": {"observed_value": "foo"},
    }
    rendered_content = get_diagnostic_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.diagnostic.observed_value",
        "value": {
            "params": {"observed_value": {"schema": {"type": "string"}, "value": "foo"}},
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$observed_value",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_diagnostic_observed_value_with_unexpected_percent(get_diagnostic_rendered_content):
    # Please note that the vast majority of Expectations are calling `Expectation._atomic_diagnostic_observed_value()`  # noqa: E501 # FIXME CoP
    # As such, the specific expectation_type used here is irrelevant and is simply used to trigger the parent class.  # noqa: E501 # FIXME CoP
    expectation_config = {
        "type": "expect_table_row_count_to_equal",
        "kwargs": {},
    }
    update_dict = {
        "expectation_config": ExpectationConfiguration(**expectation_config),
        "result": {"unexpected_percent": 10},
    }
    rendered_content = get_diagnostic_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.diagnostic.observed_value",
        "value": {
            "params": {"observed_value": {"schema": {"type": "string"}, "value": "10% unexpected"}},
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$observed_value",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
def test_atomic_diagnostic_observed_value_with_empty_result(get_diagnostic_rendered_content):
    # Please note that the vast majority of Expectations are calling `Expectation._atomic_diagnostic_observed_value()`  # noqa: E501 # FIXME CoP
    # As such, the specific expectation_type used here is irrelevant and is simply used to trigger the parent class.  # noqa: E501 # FIXME CoP
    expectation_config = {
        "type": "expect_table_row_count_to_equal",
        "kwargs": {},
    }
    update_dict = {
        "expectation_config": ExpectationConfiguration(**expectation_config),
        "result": {},
    }
    rendered_content = get_diagnostic_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.diagnostic.observed_value",
        "value": {
            "params": {"observed_value": {"schema": {"type": "string"}, "value": "--"}},
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": "$observed_value",
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
@pytest.mark.parametrize(
    "result, expected_template, expected_params",
    [
        (
            {"observed_value": "foo"},
            "$observed_value",
            {"observed_value": {"schema": {"type": "string"}, "value": "foo"}},
        ),
        (
            {"observed_value": False},
            "$observed_value",
            {"observed_value": {"schema": {"type": "boolean"}, "value": False}},
        ),
        (
            {
                "observed_value": datetime.strptime(
                    "1999-12-31T23:59:59-00:00", "%Y-%m-%dT%H:%M:%S%z"
                )
            },
            "$observed_value",
            {
                "observed_value": {
                    "schema": {"type": "datetime"},
                    "value": datetime.strptime("1999-12-31T23:59:59-00:00", "%Y-%m-%dT%H:%M:%S%z"),
                }
            },
        ),
        (
            {"observed_value": "1999-12-31 23:59:59"},
            "$observed_value",
            {"observed_value": {"schema": {"type": "datetime"}, "value": "1999-12-31 23:59:59"}},
        ),
        (
            {"observed_value": 1776},
            "$observed_value",
            {"observed_value": {"schema": {"type": "number"}, "value": 1776}},
        ),
        (
            {"unexpected_percent": 10},
            "$observed_value",
            {"observed_value": {"schema": {"type": "string"}, "value": "10% unexpected"}},
        ),
        (
            {"observed_value": ["%", "_"]},
            "$observed_value",
            {"observed_value": {"schema": {"type": "array"}, "value": ["%", "_"]}},
        ),
        (
            {"observed_value": [1, 2, 3]},
            "$observed_value",
            {"observed_value": {"schema": {"type": "array"}, "value": [1, 2, 3]}},
        ),
        # when all value types are inferrable string will end up taking precendence over
        # object otherwise object would overrride most other inference types as it
        # is permissible to any value type
        (
            {"observed_value": {"foo": "bar"}},
            "$observed_value",
            {"observed_value": {"schema": {"type": "string"}, "value": {"foo": "bar"}}},
        ),
        (
            {"observed_value": None},
            "$observed_value",
            {"observed_value": {"schema": {"type": "string"}, "value": "--"}},
        ),
        (
            {},
            "$observed_value",
            {"observed_value": {"schema": {"type": "string"}, "value": "--"}},
        ),
    ],
)
def test_atomic_diagnostic_observed_param_type_inference(
    get_diagnostic_rendered_content, result, expected_template, expected_params
):
    expectation_config = {
        "type": "expect_table_row_count_to_equal",
        "kwargs": {},
    }
    update_dict = {
        "expectation_config": ExpectationConfiguration(**expectation_config),
        "result": result,
    }
    rendered_content = get_diagnostic_rendered_content(update_dict)
    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.diagnostic.observed_value",
        "value": {
            "params": expected_params,
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": expected_template,
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
@pytest.mark.parametrize(
    "description, expected_value, observed_value, expected_result",
    [
        (
            "happy",
            ["a", "b", "c"],
            ["a", "b", "c"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("ov__2", "c", "expected"),
            ],
        ),
        (
            "disjoint",
            ["a", "b", "c"],
            ["x", "y", "z"],
            [
                ("exp__0", "a", "missing"),
                ("exp__1", "b", "missing"),
                ("exp__2", "c", "missing"),
                ("ov__0", "x", "unexpected"),
                ("ov__1", "y", "unexpected"),
                ("ov__2", "z", "unexpected"),
            ],
        ),
        (
            "transposed chars",
            ["a", "b", "c", "d"],
            ["a", "c", "b", "d"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "c", "unexpected"),
                ("ov__2", "b", "unexpected"),
                ("ov__3", "d", "expected"),
            ],
        ),
        (
            "renamed",
            ["a", "b", "c"],
            ["a", "c", "b2"],
            [
                ("ov__0", "a", ObservedValueRenderState.EXPECTED),
                ("exp__1", "b", ObservedValueRenderState.MISSING),
                ("ov__1", "c", ObservedValueRenderState.EXPECTED),
                ("ov__2", "b2", ObservedValueRenderState.UNEXPECTED),
            ],
        ),
        (
            "pair transposed",
            ["a", "b"],
            ["b", "a"],
            [
                ("ov__0", "b", ObservedValueRenderState.UNEXPECTED),
                ("ov__1", "a", ObservedValueRenderState.UNEXPECTED),
            ],
        ),
        (
            # NOTE: this is a confusing one, but it is hard to generalize
            "first unexpected and last missing",
            ["a", "b"],
            ["x", "a"],
            [
                ("ov__0", "x", ObservedValueRenderState.UNEXPECTED),
                ("exp__1", "b", ObservedValueRenderState.MISSING),
                ("ov__1", "a", ObservedValueRenderState.UNEXPECTED),
            ],
        ),
        (
            "pair second index missing",
            ["a", "b"],
            ["a", "x"],
            [
                ("ov__0", "a", ObservedValueRenderState.EXPECTED),
                ("exp__1", "b", ObservedValueRenderState.MISSING),
                ("ov__1", "x", ObservedValueRenderState.UNEXPECTED),
            ],
        ),
        (
            "empty actual",
            ["a", "b"],
            [],
            [
                ("exp__0", "a", ObservedValueRenderState.MISSING),
                ("exp__1", "b", ObservedValueRenderState.MISSING),
            ],
        ),
        (
            "one column deleted",
            ["a", "b", "c"],
            ["a", "c"],
            [
                ("ov__0", "a", ObservedValueRenderState.EXPECTED),
                ("exp__1", "b", ObservedValueRenderState.MISSING),
                ("ov__1", "c", ObservedValueRenderState.EXPECTED),
            ],
        ),
        (
            "last column deleted",
            ["a", "b", "c", "d"],
            ["a", "b", "c"],
            [
                ("ov__0", "a", ObservedValueRenderState.EXPECTED),
                ("ov__1", "b", ObservedValueRenderState.EXPECTED),
                ("ov__2", "c", ObservedValueRenderState.EXPECTED),
                ("exp__3", "d", ObservedValueRenderState.MISSING),
            ],
        ),
        (
            "empty expected",
            [],
            ["a", "b"],
            [
                ("ov__0", "a", ObservedValueRenderState.UNEXPECTED),
                ("ov__1", "b", ObservedValueRenderState.UNEXPECTED),
            ],
        ),
        (
            "one column added",
            ["a", "b"],
            ["a", "b", "c"],
            [
                ("ov__0", "a", ObservedValueRenderState.EXPECTED),
                ("ov__1", "b", ObservedValueRenderState.EXPECTED),
                ("ov__2", "c", ObservedValueRenderState.UNEXPECTED),
            ],
        ),
        (
            "missing bookends",
            ["f", "a", "b", "c", "d"],
            ["a", "b", "c"],
            [
                ("exp__0", "f", ObservedValueRenderState.MISSING),
                ("ov__0", "a", ObservedValueRenderState.EXPECTED),
                ("ov__1", "b", ObservedValueRenderState.EXPECTED),
                ("ov__2", "c", ObservedValueRenderState.EXPECTED),
                ("exp__4", "d", ObservedValueRenderState.MISSING),
            ],
        ),
        (
            "mix 2",
            ["a", "b", "c", "d"],
            ["a", "c", "d", "b", "e"],
            [
                ("ov__0", "a", ObservedValueRenderState.EXPECTED),
                ("ov__1", "c", ObservedValueRenderState.UNEXPECTED),
                ("ov__2", "d", ObservedValueRenderState.UNEXPECTED),
                ("ov__3", "b", ObservedValueRenderState.UNEXPECTED),
                ("ov__4", "e", ObservedValueRenderState.UNEXPECTED),
            ],
        ),
    ],
)
def test_expect_table_columns_to_match_ordered_list_atomic_diagnostic_observed_value(
    description,
    expected_value,
    observed_value,
    expected_result,
    get_diagnostic_rendered_content,
):
    # arrange
    x = {
        "expectation_config": ExpectationConfiguration(
            type="expect_table_columns_to_match_ordered_list",
            kwargs={"column_list": expected_value},
        ),
        "result": {"observed_value": observed_value},
    }

    expected_template_string = " ".join([f"${name}" for name, _, _ in expected_result])

    # act
    res = get_diagnostic_rendered_content(x).to_json_dict()

    # assert
    assert res["value"]["template"] == expected_template_string

    for name, val, status in expected_result:
        assert name in res["value"]["params"]
        assert res["value"]["params"][name]["value"] == val
        assert res["value"]["params"][name]["render_state"] == status


@pytest.mark.unit
@pytest.mark.parametrize(
    "value_set, result, expected_template, expected_params",
    [
        (
            ["blue", "green"],
            {"observed_value": ["blue", "red"]},
            "$ov__0 $ov__1",
            {
                "observed_value": {"schema": {"type": "array"}, "value": ["blue", "red"]},
                "ov__0": {
                    "schema": {"type": "string"},
                    "value": "blue",
                    "render_state": "expected",
                },
                "ov__1": {
                    "schema": {"type": "string"},
                    "value": "red",
                    "render_state": "unexpected",
                },
                "value_set": {"schema": {"type": "array"}, "value": ["blue", "green"]},
            },
        ),
        (
            ["blue", "green"],
            {"observed_value": ["red"]},
            "$ov__0",
            {
                "observed_value": {"schema": {"type": "array"}, "value": ["red"]},
                "ov__0": {
                    "schema": {"type": "string"},
                    "value": "red",
                    "render_state": "unexpected",
                },
                "value_set": {"schema": {"type": "array"}, "value": ["blue", "green"]},
            },
        ),
    ],
)
def test_expect_column_most_common_value_to_be_in_set_atomic_diagnostic_observed_value(
    get_diagnostic_rendered_content, value_set, result, expected_template, expected_params
):
    # arrange
    x = {
        "expectation_config": ExpectationConfiguration(
            type="expect_column_most_common_value_to_be_in_set",
            kwargs={
                "column": "color",
                "value_set": value_set,
                # ties_okay parameter does not affect observed value rendering
                # the Expectation can pass and still have values with
                # render_state "unexpected" in the observed value set
                # if ties_okay is set to True
            },
        ),
        "result": result,
    }

    # act
    rendered_content = get_diagnostic_rendered_content(x)

    # assert
    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.diagnostic.observed_value",
        "value": {
            "params": expected_params,
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": expected_template,
        },
        "value_type": "StringValueType",
    }


@pytest.mark.unit
@pytest.mark.parametrize(
    "description, value_set, observed_value, expected_result",
    [
        (
            "complete set",
            ["a", "b", "c"],
            ["a", "b", "c"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("ov__2", "c", "expected"),
            ],
        ),
        (
            "empty input",
            [],
            ["a", "b", "c"],
            [
                ("ov__0", "a", "unexpected"),
                ("ov__1", "b", "unexpected"),
                ("ov__2", "c", "unexpected"),
            ],
        ),
        (
            "empty observed",
            ["a", "b", "c"],
            [],
            [],
        ),
        (
            "empty input and observed",
            [],
            [],
            [],
        ),
        (
            "subset observed",
            ["a", "b", "c"],
            ["a", "b"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
            ],
        ),
        (
            "superset observed",
            ["a", "b", "c"],
            ["a", "b", "c", "d"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("ov__2", "c", "expected"),
                ("ov__3", "d", "unexpected"),
            ],
        ),
    ],
)
def test_expect_column_distinct_values_to_be_in_set_atomic_diagnostic_observed_value(
    description,
    value_set,
    observed_value,
    expected_result,
    get_diagnostic_rendered_content,
):
    # arrange
    x = {
        "expectation_config": ExpectationConfiguration(
            type="expect_column_distinct_values_to_be_in_set",
            kwargs={"value_set": value_set},
        ),
        "result": {"observed_value": observed_value},
    }

    expected_template_string = " ".join([f"${name}" for name, _, _ in expected_result])

    # act
    res = get_diagnostic_rendered_content(x).to_json_dict()

    # assert
    assert res["value"]["template"] == expected_template_string

    for name, val, status in expected_result:
        assert name in res["value"]["params"]
        assert res["value"]["params"][name]["value"] == val
        assert res["value"]["params"][name]["render_state"] == status


@pytest.mark.unit
@pytest.mark.parametrize(
    "description, value_set, observed_value, expected_result",
    [
        (
            "complete set",
            ["a", "b", "c"],
            ["a", "b", "c"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("ov__2", "c", "expected"),
            ],
        ),
        (
            "empty input",
            [],
            ["a", "b", "c"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("ov__2", "c", "expected"),
            ],
        ),
        (
            "empty observed",
            ["a", "b", "c"],
            [],
            [
                ("exp__0", "a", "missing"),
                ("exp__1", "b", "missing"),
                ("exp__2", "c", "missing"),
            ],
        ),
        (
            "empty input and observed",
            [],
            [],
            [],
        ),
        (
            "subset observed",
            ["a", "b", "c"],
            ["a", "b"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("exp__2", "c", "missing"),
            ],
        ),
        (
            "superset observed",
            ["a", "b", "c"],
            ["a", "b", "c", "d"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("ov__2", "c", "expected"),
                ("ov__3", "d", "expected"),
            ],
        ),
    ],
)
def test_expect_column_distinct_values_to_contain_set_atomic_diagnostic_observed_value(
    description,
    value_set,
    observed_value,
    expected_result,
    get_diagnostic_rendered_content,
):
    # arrange
    x = {
        "expectation_config": ExpectationConfiguration(
            type="expect_column_distinct_values_to_contain_set",
            kwargs={"value_set": value_set},
        ),
        "result": {"observed_value": observed_value},
    }

    expected_template_string = " ".join([f"${name}" for name, _, _ in expected_result])

    # act
    res = get_diagnostic_rendered_content(x).to_json_dict()

    # assert
    assert res["value"]["template"] == expected_template_string

    for name, val, status in expected_result:
        assert name in res["value"]["params"]
        assert res["value"]["params"][name]["value"] == val
        assert res["value"]["params"][name]["render_state"] == status


def _create_result_details_from_expected_result(
    expected_result: Tuple[str, str, str],
) -> Optional[Dict[str, Any]]:
    unexpected = []
    missing = []
    for _, col, state in expected_result:
        if state == "unexpected":
            unexpected.append(col)
        elif state == "missing":
            missing.append(col)
    if not unexpected and not missing:
        return None
    details = {"mismatched": {}}
    if unexpected:
        details["mismatched"]["unexpected"] = unexpected
    if missing:
        details["mismatched"]["missing"] = missing
    return details


@pytest.mark.unit
@pytest.mark.parametrize(
    "description, value_set, observed_value, expected_result",
    [
        (
            "complete set",
            ["a", "b", "c"],
            ["a", "b", "c"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("ov__2", "c", "expected"),
            ],
        ),
        (
            "empty input",
            [],
            ["a", "b", "c"],
            [
                ("ov__0", "a", "unexpected"),
                ("ov__1", "b", "unexpected"),
                ("ov__2", "c", "unexpected"),
            ],
        ),
        (
            "empty observed",
            ["a", "b", "c"],
            [],
            [
                ("exp__0", "a", "missing"),
                ("exp__1", "b", "missing"),
                ("exp__2", "c", "missing"),
            ],
        ),
        (
            "empty input and observed",
            [],
            [],
            [],
        ),
        (
            "subset observed",
            ["a", "b", "c"],
            ["a", "b"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("exp__2", "c", "missing"),
            ],
        ),
        (
            "superset observed",
            ["a", "b", "c"],
            ["a", "b", "c", "d"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("ov__2", "c", "expected"),
                ("ov__3", "d", "unexpected"),
            ],
        ),
        (
            "superset observed 2",
            ["a", "b", "c"],
            ["a", "d", "b", "c"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "d", "unexpected"),
                ("ov__2", "b", "expected"),
                ("ov__3", "c", "expected"),
            ],
        ),
    ],
)
def test_expect_column_distinct_values_to_equal_set_atomic_diagnostic_observed_value(
    description,
    value_set,
    observed_value,
    expected_result,
    get_diagnostic_rendered_content,
):
    # arrange
    x = {
        "expectation_config": ExpectationConfiguration(
            type="expect_column_distinct_values_to_equal_set",
            kwargs={"value_set": value_set},
        ),
        "result": {"observed_value": observed_value},
    }

    expected_template_string = " ".join([f"${name}" for name, _, _ in expected_result])

    # act
    res = get_diagnostic_rendered_content(x).to_json_dict()

    # assert
    assert res["value"]["template"] == expected_template_string

    for name, val, status in expected_result:
        assert name in res["value"]["params"]
        assert res["value"]["params"][name]["value"] == val
        assert res["value"]["params"][name]["render_state"] == status


@pytest.mark.unit
@pytest.mark.parametrize(
    "description, column_set, observed_value, expected_result",
    [
        (
            "complete set",
            ["a", "b", "c"],
            ["a", "b", "c"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("ov__2", "c", "expected"),
            ],
        ),
        (
            "empty input",
            [],
            ["a", "b", "c"],
            [
                ("ov__0", "a", "unexpected"),
                ("ov__1", "b", "unexpected"),
                ("ov__2", "c", "unexpected"),
            ],
        ),
        (
            "empty observed",
            ["a", "b", "c"],
            [],
            [
                ("exp__0", "a", "missing"),
                ("exp__1", "b", "missing"),
                ("exp__2", "c", "missing"),
            ],
        ),
        (
            "empty input and observed",
            [],
            [],
            [],
        ),
        (
            "subset observed",
            ["a", "b", "c"],
            ["a", "b"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("exp__2", "c", "missing"),
            ],
        ),
        (
            "superset observed",
            ["a", "b", "c"],
            ["a", "b", "c", "d"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "b", "expected"),
                ("ov__2", "c", "expected"),
                ("ov__3", "d", "unexpected"),
            ],
        ),
        (
            "superset observed 2",
            ["a", "b", "c"],
            ["a", "d", "b", "c"],
            [
                ("ov__0", "a", "expected"),
                ("ov__1", "d", "unexpected"),
                ("ov__2", "b", "expected"),
                ("ov__3", "c", "expected"),
            ],
        ),
    ],
)
def test_expect_table_columns_to_match_set_atomic_diagnostic_observed_value(
    description,
    column_set,
    observed_value,
    expected_result,
    get_diagnostic_rendered_content,
):
    # arrange
    x = {
        "expectation_config": ExpectationConfiguration(
            type="expect_table_columns_to_match_set",
            kwargs={"column_set": column_set},
        ),
        "result": {"observed_value": observed_value},
    }
    details = _create_result_details_from_expected_result(expected_result)
    if details:
        x["result"]["details"] = details

    expected_template_string = " ".join([f"${name}" for name, _, _ in expected_result])

    # act
    res = get_diagnostic_rendered_content(x).to_json_dict()

    # assert
    assert res["value"]["template"] == expected_template_string

    for name, val, status in expected_result:
        assert name in res["value"]["params"]
        assert res["value"]["params"][name]["value"] == val
        assert res["value"]["params"][name]["render_state"] == status


@pytest.mark.unit
@pytest.mark.parametrize(
    "observed_value, expected_template_string",
    [
        (
            0,
            "$observed_value unexpected rows",
        ),
        (
            1,
            "$observed_value unexpected row",
        ),
        (
            100000,
            "$observed_value unexpected rows",
        ),
    ],
)
def test_unexpected_rows_expectation_atomic_diagnostic_observed_value(
    observed_value,
    expected_template_string,
    get_diagnostic_rendered_content,
):
    # arrange
    x = {
        "expectation_config": ExpectationConfiguration(
            type="unexpected_rows_expectation",
            kwargs={"description": "my description", "unexpected_rows_query": "valid query"},
        ),
        "result": {"observed_value": observed_value},
    }

    # act
    res = get_diagnostic_rendered_content(x).to_json_dict()

    # assert
    assert res["value"]["template"] == expected_template_string


@pytest.mark.unit
def test_unexpected_rows_expectation_atomic_diagnostic_observed_value_when_description_present(
    get_diagnostic_rendered_content,
):
    """Fixes regression where description overwrote the template"""

    x = {
        "expectation_config": ExpectationConfiguration(
            type="unexpected_rows_expectation",
            kwargs={"description": "my description", "unexpected_rows_query": "valid query"},
            description="plz ignore me",
        ),
        "result": {"observed_value": 123},
    }

    res = get_diagnostic_rendered_content(x).to_json_dict()

    assert res["value"]["template"] == "$observed_value unexpected rows"


@pytest.mark.unit
def test_atomic_prescriptive_summary_with_description(
    get_prescriptive_rendered_content,
):
    description = "I should overwite"
    update_dict = {
        "type": "expect_column_distinct_values_to_be_in_set",
        "description": description,
        "kwargs": {
            "column": "my_column",
            "value_set": [1, 2, 3],
        },
    }
    rendered_content = get_prescriptive_rendered_content(update_dict)

    res = rendered_content.to_json_dict()
    pprint(res)
    assert res == {
        "name": "atomic.prescriptive.summary",
        "value": {
            "params": {
                "column": {"schema": {"type": "string"}, "value": "my_column"},
                "v__0": {"schema": {"type": "number"}, "value": 1},
                "v__1": {"schema": {"type": "number"}, "value": 2},
                "v__2": {"schema": {"type": "number"}, "value": 3},
                "value_set": {"schema": {"type": "array"}, "value": [1, 2, 3]},
            },
            "schema": {"type": "com.superconductive.rendered.string"},
            "template": description,
        },
        "value_type": "StringValueType",
    }



from typing import Sequence

import pandas as pd
import pytest

from great_expectations.metrics.query.row_count import QueryRowCount, QueryRowCountResult
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import (
    BigQueryDatasourceTestConfig,
    DatabricksDatasourceTestConfig,
    DataSourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SparkFilesystemCsvDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

SPARK_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    SparkFilesystemCsvDatasourceTestConfig(),
]

SQL_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(),
    DatabricksDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
]

DATA_FRAME = pd.DataFrame(
    {
        "id": [1, 2, 3, 4],
        "name": ["A", "B", "C", "D"],
    },
)


class TestQueryRowCount:
    @parameterize_batch_for_data_sources(
        data_source_configs=SPARK_DATA_SOURCES,
        data=DATA_FRAME,
    )
    def test_success_spark(self, batch_for_datasource) -> None:
        batch = batch_for_datasource
        metric = QueryRowCount(query="SELECT * FROM {batch} WHERE id > 0")
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, QueryRowCountResult)
        assert metric_result.value == 4

    @parameterize_batch_for_data_sources(
        data_source_configs=SQL_DATA_SOURCES,
        data=DATA_FRAME,
    )
    @pytest.mark.parametrize(
        ["query", "result_value"],
        [("SELECT * FROM {batch} WHERE id > 0", 4), ("SELECT * FROM {batch} WHERE id <= 2", 2)],
    )
    def test_success_sql(self, batch_for_datasource, query, result_value) -> None:
        batch = batch_for_datasource
        metric = QueryRowCount(query=query)
        metric_result = batch.compute_metrics(metric)
        assert isinstance(metric_result, QueryRowCountResult)
        assert metric_result.value == result_value



import os

import boto3
import pytest
from moto import mock_sns
from pytest_mock import MockerFixture

from great_expectations.core import ExpectationSuiteValidationResult, RunIdentifier
from great_expectations.data_context import set_context
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext
from great_expectations.data_context.types.resource_identifiers import (
    BatchIdentifier,
    ExpectationSuiteIdentifier,
    GXCloudIdentifier,
    ValidationResultIdentifier,
)


@pytest.fixture(scope="module")
def validation_result_suite_ge_cloud_id():
    return "bfe7dc64-5320-49b0-91c1-2e8029e06c4d"


@pytest.fixture(scope="module")
def validation_result_suite():
    return ExpectationSuiteValidationResult(
        results=[],
        success=True,
        suite_name="empty_suite",
        statistics={
            "evaluated_expectations": 0,
            "successful_expectations": 0,
            "unsuccessful_expectations": 0,
            "success_percent": None,
        },
        meta={
            "great_expectations_version": "v0.8.0__develop",
            "expectation_suite_name": "asset.default",
            "run_id": "test_100",
        },
    )


@pytest.fixture(scope="module")
def validation_result_suite_ge_cloud_identifier(validation_result_suite_ge_cloud_id):
    return GXCloudIdentifier(
        resource_type=GXCloudRESTResource.CHECKPOINT,
        id=validation_result_suite_ge_cloud_id,
    )


@pytest.fixture(scope="module")
def validation_result_suite_with_ge_cloud_id(validation_result_suite_ge_cloud_id):
    return ExpectationSuiteValidationResult(
        results=[],
        success=True,
        suite_name="empty_suite",
        statistics={
            "evaluated_expectations": 0,
            "successful_expectations": 0,
            "unsuccessful_expectations": 0,
            "success_percent": None,
        },
        meta={
            "great_expectations_version": "v0.8.0__develop",
            "expectation_suite_name": "asset.default",
            "run_id": "test_100",
        },
        id=validation_result_suite_ge_cloud_id,
    )


@pytest.fixture(scope="module")
def validation_result_suite_id():
    return ValidationResultIdentifier(
        expectation_suite_identifier=ExpectationSuiteIdentifier("asset.default"),
        run_id=RunIdentifier(run_name="test_100"),
        batch_identifier="1234",
    )


@pytest.fixture(scope="module")
def validation_result_suite_extended_id():
    return ValidationResultIdentifier(
        expectation_suite_identifier=ExpectationSuiteIdentifier("asset.default"),
        run_id=RunIdentifier(run_name="test_100", run_time="Tue May 08 15:14:45 +0800 2012"),
        batch_identifier=BatchIdentifier(batch_identifier="1234", data_asset_name="asset"),
    )


@pytest.fixture
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"


@pytest.fixture
def sns(aws_credentials):
    with mock_sns():
        conn = boto3.client("sns")
        yield conn


@pytest.fixture
def mock_context(mocker: MockerFixture):
    context = mocker.MagicMock(spec=AbstractDataContext)
    set_context(context)
    return context



import pandas as pd

from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.metrics.column.distinct_values import (
    ColumnDistinctValues,
    ColumnDistinctValuesResult,
)
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import PANDAS_DATA_SOURCES, SPARK_DATA_SOURCES, SQL_DATA_SOURCES

COLUMN_NAME = "whatevs"
DATA_FRAME = pd.DataFrame(
    {
        COLUMN_NAME: ["a", "b", "c", "c", "c", "c", "c", "c", "c", "c", None, None],
    },
)


def get_pandas_data_sources():
    return PANDAS_DATA_SOURCES


def get_non_pandas_data_sources():
    return SPARK_DATA_SOURCES + SQL_DATA_SOURCES


class TestColumnDistinctValues:
    @parameterize_batch_for_data_sources(
        data_source_configs=get_pandas_data_sources(),
        data=DATA_FRAME,
    )
    def test_distinct_values_pandas(self, batch_for_datasource: Batch) -> None:
        metric = ColumnDistinctValues(column=COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnDistinctValuesResult)
        # For pandas, we expect the null values to be included
        assert len(metric_result.value) == 4  # a, b, c, and null
        assert "a" in metric_result.value
        assert "b" in metric_result.value
        assert "c" in metric_result.value
        # Check for either None or nan depending on source
        assert any(pd.isna(val) for val in metric_result.value)

    @parameterize_batch_for_data_sources(
        data_source_configs=get_non_pandas_data_sources(),
        data=DATA_FRAME,
    )
    def test_distinct_values_non_pandas(self, batch_for_datasource: Batch) -> None:
        metric = ColumnDistinctValues(column=COLUMN_NAME)
        metric_result = batch_for_datasource.compute_metrics(metric)

        assert isinstance(metric_result, ColumnDistinctValuesResult)
        # For SQL and Spark, we expect only the non-null values
        assert metric_result.value == {"a", "b", "c"}



from __future__ import annotations

import datetime
from typing import Optional, Union

import pandas as pd
from dateutil.parser import parse

from great_expectations.compatibility.pyspark import functions as F
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.map_metric_provider import (
    ColumnMapMetricProvider,
    column_condition_partial,
)


class InvalidColumnTypeError(ValueError):
    def __init__(self, column_type: str):
        message = f"ColumnValuesBetween metrics cannot be computed on column of type {column_type}."
        super().__init__(message)


class ColumnValuesBetween(ColumnMapMetricProvider):
    condition_metric_name = "column_values.between"
    condition_value_keys = (
        "min_value",
        "max_value",
        "strict_min",
        "strict_max",
    )

    @column_condition_partial(engine=PandasExecutionEngine)
    def _pandas(  # noqa: C901 # FIXME CoP
        cls,
        column,
        min_value=None,
        max_value=None,
        strict_min=None,
        strict_max=None,
        **kwargs,
    ):
        if min_value is None and max_value is None:
            raise ValueError("min_value and max_value cannot both be None")  # noqa: TRY003 # FIXME CoP

        temp_column = column

        if min_value is not None and max_value is not None and min_value > max_value:
            raise ValueError("min_value cannot be greater than max_value")  # noqa: TRY003 # FIXME CoP

        # Use a vectorized approach for native numpy dtypes
        if column.dtype in [int, float]:
            return cls._pandas_vectorized(temp_column, min_value, max_value, strict_min, strict_max)
        elif isinstance(column.dtype, pd.DatetimeTZDtype) or pd.api.types.is_datetime64_ns_dtype(
            column.dtype
        ):
            if min_value is not None and isinstance(min_value, str):
                min_value = parse(min_value)

            if max_value is not None and isinstance(max_value, str):
                max_value = parse(max_value)

            return cls._pandas_vectorized(temp_column, min_value, max_value, strict_min, strict_max)

        def is_between(val):  # noqa: C901, PLR0911, PLR0912 # FIXME CoP
            # TODO Might be worth explicitly defining comparisons between types (for example, between strings and ints).  # noqa: E501 # FIXME CoP
            # Ensure types can be compared since some types in Python 3 cannot be logically compared.  # noqa: E501 # FIXME CoP
            # print type(val), type(min_value), type(max_value), val, min_value, max_value

            if type(val) is None:
                return False

            if min_value is not None and max_value is not None:
                # Type of column values is either string or specific rich type (or "None").  In all cases, type of  # noqa: E501 # FIXME CoP
                # column must match type of constant being compared to column value (otherwise, error is raised).  # noqa: E501 # FIXME CoP
                if (isinstance(val, str) != isinstance(min_value, str)) or (
                    isinstance(val, str) != isinstance(max_value, str)
                ):
                    raise TypeError(  # noqa: TRY003 # FIXME CoP
                        "Column values, min_value, and max_value must either be None or of the same type."  # noqa: E501 # FIXME CoP
                    )

                if strict_min and strict_max:
                    return (val > min_value) and (val < max_value)

                if strict_min:
                    return (val > min_value) and (val <= max_value)

                if strict_max:
                    return (val >= min_value) and (val < max_value)

                return (val >= min_value) and (val <= max_value)

            elif min_value is None and max_value is not None:
                # Type of column values is either string or specific rich type (or "None").  In all cases, type of  # noqa: E501 # FIXME CoP
                # column must match type of constant being compared to column value (otherwise, error is raised).  # noqa: E501 # FIXME CoP
                if isinstance(val, str) != isinstance(max_value, str):
                    raise TypeError(  # noqa: TRY003 # FIXME CoP
                        "Column values, min_value, and max_value must either be None or of the same type."  # noqa: E501 # FIXME CoP
                    )

                if strict_max:
                    return val < max_value

                return val <= max_value

            elif min_value is not None and max_value is None:
                # Type of column values is either string or specific rich type (or "None").  In all cases, type of  # noqa: E501 # FIXME CoP
                # column must match type of constant being compared to column value (otherwise, error is raised).  # noqa: E501 # FIXME CoP
                if isinstance(val, str) != isinstance(min_value, str):
                    raise TypeError(  # noqa: TRY003 # FIXME CoP
                        "Column values, min_value, and max_value must either be None or of the same type."  # noqa: E501 # FIXME CoP
                    )

                if strict_min:
                    return val > min_value

                return val >= min_value

            else:
                return False

        return temp_column.map(is_between)

    @classmethod
    def _pandas_vectorized(  # noqa: C901, PLR0911 # FIXME CoP
        cls,
        column: pd.Series,
        min_value: Optional[Union[int, float, datetime.datetime]],
        max_value: Optional[Union[int, float, datetime.datetime]],
        strict_min: bool,
        strict_max: bool,
    ):
        if min_value is None and max_value is None:
            raise ValueError("min_value and max_value cannot both be None")  # noqa: TRY003 # FIXME CoP

        if min_value is None:
            if strict_max:
                return column < max_value
            else:
                return column <= max_value

        if max_value is None:
            if strict_min:
                return min_value < column
            else:
                return min_value <= column

        if strict_min and strict_max:
            return (min_value < column) & (column < max_value)
        elif strict_min:
            return (min_value < column) & (column <= max_value)
        elif strict_max:
            return (min_value <= column) & (column < max_value)
        else:
            return (min_value <= column) & (column <= max_value)

    @column_condition_partial(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(  # noqa: C901, PLR0911 # FIXME CoP
        cls,
        column,
        min_value=None,
        max_value=None,
        strict_min=None,
        strict_max=None,
        **kwargs,
    ):
        if min_value is not None and max_value is not None and min_value > max_value:
            raise ValueError("min_value cannot be greater than max_value")  # noqa: TRY003 # FIXME CoP

        if min_value is None and max_value is None:
            raise ValueError("min_value and max_value cannot both be None")  # noqa: TRY003 # FIXME CoP

        # Check that the generated SQL won't raise an error.
        # ColumnValuesBetween metrics only work on numbers/dates,
        # so we check for common string and boolean types.

        # Retrieve column types from metrics.
        metrics = kwargs.get("_metrics", {})
        column_types = metrics.get("table.column_types", [])

        # Map column names to their types as strings.
        type_by_column = {ct.get("name"): str(ct.get("type", "")) for ct in column_types}
        column_type = type_by_column.get(column.name)

        INVALID_COLUMN_TYPES = (
            "VARCHAR",
            "CHAR",
            "NVARCHAR",
            "NCHAR",
            "TEXT",
            "STRING",
            "BOOLEAN",
            "BOOL",
            "BIT",
            "TINYTEXT",
            "MEDIUMTEXT",
            "LONGTEXT",
        )
        if column_type and column_type.upper().startswith(INVALID_COLUMN_TYPES):
            raise InvalidColumnTypeError(column_type=column_type)

        if min_value is None:
            if strict_max:
                return column < sa.literal(max_value)

            return column <= sa.literal(max_value)

        elif max_value is None:
            if strict_min:
                return column > sa.literal(min_value)

            return column >= sa.literal(min_value)

        else:
            if strict_min and strict_max:
                return sa.and_(
                    column > sa.literal(min_value),
                    column < sa.literal(max_value),
                )

            if strict_min:
                return sa.and_(
                    column > sa.literal(min_value),
                    column <= sa.literal(max_value),
                )

            if strict_max:
                return sa.and_(
                    column >= sa.literal(min_value),
                    column < sa.literal(max_value),
                )

            return sa.and_(
                column >= sa.literal(min_value),
                column <= sa.literal(max_value),
            )

    @column_condition_partial(engine=SparkDFExecutionEngine)
    def _spark(  # noqa: C901, PLR0911 # FIXME CoP
        cls,
        column,
        min_value=None,
        max_value=None,
        strict_min=None,
        strict_max=None,
        **kwargs,
    ):
        if min_value is not None and max_value is not None and min_value > max_value:
            raise ValueError("min_value cannot be greater than max_value")  # noqa: TRY003 # FIXME CoP

        if min_value is None and max_value is None:
            raise ValueError("min_value and max_value cannot both be None")  # noqa: TRY003 # FIXME CoP

        if min_value is None:
            if strict_max:
                return column < F.lit(max_value)

            return column <= F.lit(max_value)

        elif max_value is None:
            if strict_min:
                return column > F.lit(min_value)

            return column >= F.lit(min_value)

        else:
            if strict_min and strict_max:
                return (column > F.lit(min_value)) & (column < F.lit(max_value))

            if strict_min:
                return (column > F.lit(min_value)) & (column <= F.lit(max_value))

            if strict_max:
                return (column >= F.lit(min_value)) & (column < F.lit(max_value))

            return (column >= F.lit(min_value)) & (column <= F.lit(max_value))



from __future__ import annotations

from great_expectations.compatibility.not_imported import NotImported

TRINO_NOT_IMPORTED = NotImported(
    "trino connection components are not installed, please 'pip install trino'"
)

try:
    import trino
except ImportError:
    trino = TRINO_NOT_IMPORTED

try:
    from trino.sqlalchemy import datatype as trinotypes
except (ImportError, AttributeError):
    trinotypes = TRINO_NOT_IMPORTED

try:
    from trino.sqlalchemy import dialect as trinodialect
except (ImportError, AttributeError):
    trinodialect = TRINO_NOT_IMPORTED

try:
    import trino.drivers as trinodrivers
except (ImportError, AttributeError):
    trinodrivers = TRINO_NOT_IMPORTED

try:
    import trino.exceptions as trinoexceptions
except (ImportError, AttributeError):
    trinoexceptions = TRINO_NOT_IMPORTED



from __future__ import annotations

import logging
import uuid
from pathlib import Path
from typing import Any, Dict, Tuple

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility import sqlalchemy
from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context.store.store_backend import StoreBackend
from great_expectations.util import (
    filter_properties_dict,
    get_sqlalchemy_url,
    import_make_url,
)

if sa:
    make_url = import_make_url()
    SQLAlchemyError = sqlalchemy.SQLAlchemyError


logger = logging.getLogger(__name__)


class DatabaseStoreBackend(StoreBackend):
    def __init__(  # noqa: C901, PLR0912, PLR0913 # FIXME CoP
        self,
        table_name,
        key_columns,
        fixed_length_key=True,
        credentials=None,
        url=None,
        connection_string=None,
        engine=None,
        store_name=None,
        suppress_store_backend_id=False,
        manually_initialize_store_backend_id: str = "",
        **kwargs,
    ) -> None:
        super().__init__(
            fixed_length_key=fixed_length_key,
            suppress_store_backend_id=suppress_store_backend_id,
            manually_initialize_store_backend_id=manually_initialize_store_backend_id,
            store_name=store_name,
        )
        if not sa:
            raise gx_exceptions.DataContextError(  # noqa: TRY003 # FIXME CoP
                "ModuleNotFoundError: No module named 'sqlalchemy'"
            )

        if not self.fixed_length_key:
            raise gx_exceptions.InvalidConfigError(  # noqa: TRY003 # FIXME CoP
                "DatabaseStoreBackend requires use of a fixed-length-key"
            )

        self._schema_name = None
        self._credentials = credentials
        self._connection_string = connection_string
        self._url = url

        if engine is not None:
            if credentials is not None:
                logger.warning(
                    "Both credentials and engine were provided during initialization of SqlAlchemyExecutionEngine. "  # noqa: E501 # FIXME CoP
                    "Ignoring credentials."
                )
            self.engine = engine
        elif credentials is not None:
            self.engine = self._build_engine(credentials=credentials, **kwargs)
        elif connection_string is not None:
            self.engine = sa.create_engine(connection_string, **kwargs)
        elif url is not None:
            parsed_url = make_url(url)
            self.drivername = parsed_url.drivername
            self.engine = sa.create_engine(url, **kwargs)
        else:
            raise gx_exceptions.InvalidConfigError(  # noqa: TRY003 # FIXME CoP
                "Credentials, url, connection_string, or an engine are required for a DatabaseStoreBackend."  # noqa: E501 # FIXME CoP
            )

        meta = sa.MetaData(schema=self._schema_name)
        self.key_columns = key_columns
        # Dynamically construct a SQLAlchemy table with the name and column names we'll use
        cols = []
        for column_ in key_columns:
            if column_ == "value":
                raise gx_exceptions.InvalidConfigError(  # noqa: TRY003 # FIXME CoP
                    "'value' cannot be used as a key_element name"
                )
            cols.append(sa.Column(column_, sa.String, primary_key=True))
        cols.append(sa.Column("value", sa.String))
        try:
            table = sa.Table(table_name, meta, autoload_with=self.engine)
            # We do a "light" check: if the columns' names match, we will proceed, otherwise, create the table  # noqa: E501 # FIXME CoP
            if {str(col.name).lower() for col in table.columns} != (set(key_columns) | {"value"}):
                raise gx_exceptions.StoreBackendError(  # noqa: TRY003 # FIXME CoP
                    f"Unable to use table {table_name}: it exists, but does not have the expected schema."  # noqa: E501 # FIXME CoP
                )
        except sqlalchemy.NoSuchTableError:
            table = sa.Table(table_name, meta, *cols)
            try:
                if self._schema_name:
                    with self.engine.begin() as connection:
                        connection.execute(
                            sa.text(f"CREATE SCHEMA IF NOT EXISTS {self._schema_name};")
                        )
                meta.create_all(self.engine)
            except SQLAlchemyError as e:
                raise gx_exceptions.StoreBackendError(  # noqa: TRY003 # FIXME CoP
                    f"Unable to connect to table {table_name} because of an error. It is possible your table needs to be migrated to a new schema.  SqlAlchemyError: {e!s}"  # noqa: E501 # FIXME CoP
                )
        self._table = table
        # Initialize with store_backend_id
        self._store_backend_id = None
        self._store_backend_id = self.store_backend_id

        # Gather the call arguments of the present function (include the "module_name" and add the "class_name"), filter  # noqa: E501 # FIXME CoP
        # out the Falsy values, and set the instance "_config" variable equal to the resulting dictionary.  # noqa: E501 # FIXME CoP
        self._config = {
            "table_name": table_name,
            "key_columns": key_columns,
            "fixed_length_key": fixed_length_key,
            "credentials": credentials,
            "url": url,
            "connection_string": connection_string,
            "engine": engine,
            "store_name": store_name,
            "suppress_store_backend_id": suppress_store_backend_id,
            "manually_initialize_store_backend_id": manually_initialize_store_backend_id,
            "module_name": self.__class__.__module__,
            "class_name": self.__class__.__name__,
        }
        self._config.update(kwargs)
        filter_properties_dict(properties=self._config, clean_falsy=True, inplace=True)

    @property
    @override
    def store_backend_id(self) -> str:
        """
        Create a store_backend_id if one does not exist, and return it if it exists
        Ephemeral store_backend_id for database_store_backend until there is a place to store metadata
        Returns:
            store_backend_id which is a UUID(version=4)
        """  # noqa: E501 # FIXME CoP

        if not self._store_backend_id:
            store_id = (
                self._manually_initialize_store_backend_id
                if self._manually_initialize_store_backend_id
                else str(uuid.uuid4())
            )
            self._store_backend_id = f"{self.STORE_BACKEND_ID_PREFIX}{store_id}"
        return self._store_backend_id.replace(self.STORE_BACKEND_ID_PREFIX, "")

    def _build_engine(self, credentials, **kwargs) -> "sa.engine.Engine":  # noqa: UP037 # FIXME CoP
        """
        Using a set of given credentials, constructs an Execution Engine , connecting to a database using a URL or a
        private key path.
        """  # noqa: E501 # FIXME CoP
        # Update credentials with anything passed during connection time
        drivername = credentials.pop("drivername")
        create_engine_kwargs = kwargs
        self._schema_name = credentials.pop("schema", None)
        connect_args = credentials.pop("connect_args", None)
        if connect_args:
            create_engine_kwargs["connect_args"] = connect_args

        if "private_key_path" in credentials:
            options, create_engine_kwargs = self._get_sqlalchemy_key_pair_auth_url(
                drivername, credentials
            )
        else:
            options = get_sqlalchemy_url(drivername, **credentials)

        self.drivername = drivername

        engine = sa.create_engine(options, **create_engine_kwargs)
        return engine

    @staticmethod
    def _get_sqlalchemy_key_pair_auth_url(drivername: str, credentials: dict) -> Tuple["URL", Dict]:  # type: ignore[name-defined]  # noqa F821
        """
        Utilizing a private key path and a passphrase in a given credentials dictionary, attempts to encode the provided
        values into a private key. If passphrase is incorrect, this will fail and an exception is raised.

        Args:
            drivername(str) - The name of the driver class
            credentials(dict) - A dictionary of database credentials used to access the database

        Returns:
            a tuple consisting of a url with the serialized key-pair authentication, and a dictionary of engine kwargs.
        """  # noqa: E501 # FIXME CoP
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives import serialization

        private_key_path = credentials.pop("private_key_path")
        private_key_passphrase = credentials.pop("private_key_passphrase")

        with Path(private_key_path).expanduser().resolve().open(mode="rb") as key:
            try:
                p_key = serialization.load_pem_private_key(
                    key.read(),
                    password=private_key_passphrase.encode() if private_key_passphrase else None,
                    backend=default_backend(),
                )
            except ValueError as e:
                if "incorrect password" in str(e).lower():
                    raise gx_exceptions.DatasourceKeyPairAuthBadPassphraseError(
                        datasource_name="SqlAlchemyDatasource",
                        message="Decryption of key failed, was the passphrase incorrect?",
                    ) from e
                else:
                    raise e  # noqa: TRY201 # FIXME CoP
        pkb = p_key.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

        credentials_driver_name = credentials.pop("drivername", None)
        create_engine_kwargs = {"connect_args": {"private_key": pkb}}
        return (
            get_sqlalchemy_url(drivername or credentials_driver_name, **credentials),
            create_engine_kwargs,
        )

    def _get(self, key):  # type: ignore[explicit-override] # FIXME
        sel = (
            sa.select(sa.column("value"))
            .select_from(self._table)
            .where(
                sa.and_(
                    *(
                        getattr(self._table.columns, key_col) == val
                        for key_col, val in zip(self.key_columns, key)
                    )
                )
            )
        )
        try:
            with self.engine.begin() as connection:
                row = connection.execute(sel).fetchone()[0]
            return row
        except (IndexError, SQLAlchemyError) as e:
            logger.debug(f"Error fetching value: {e!s}")
            raise gx_exceptions.StoreError(f"Unable to fetch value for key: {key!s}")  # noqa: TRY003 # FIXME CoP

    @override
    def _get_all(self) -> list[Any]:
        raise NotImplementedError

    @override
    def _set(self, key, value, allow_update=True, **kwargs) -> None:
        cols = {k: v for (k, v) in zip(self.key_columns, key)}
        cols["value"] = value

        if allow_update:
            if self.has_key(key):
                ins = (
                    self._table.update()
                    .where(getattr(self._table.columns, self.key_columns[0]) == key[0])
                    .values(**cols)
                )
            else:
                ins = self._table.insert().values(**cols)  # type: ignore[assignment] # FIXME CoP
        else:
            ins = self._table.insert().values(**cols)  # type: ignore[assignment] # FIXME CoP

        try:
            with self.engine.begin() as connection:
                connection.execute(ins)
        except sqlalchemy.IntegrityError as e:
            if self._get(key) == value:
                logger.info(f"Key {key!s} already exists with the same value.")
            else:
                raise gx_exceptions.StoreBackendError(  # noqa: TRY003 # FIXME CoP
                    f"Integrity error {e!s} while trying to store key"
                )

    @override
    def _move(self) -> None:  # type: ignore[override] # FIXME CoP
        raise NotImplementedError

    @override
    def get_url_for_key(self, key, protocol=None) -> str:
        url = self._convert_engine_and_key_to_url(key)
        return url

    def _convert_engine_and_key_to_url(self, key):
        # SqlAlchemy engine URL is formatted in the following way
        # postgresql://postgres:password@localhost:5433/work
        # [engine]://[username]:[password]@[host]:[port]/[db_name]

        # which contains information like username and password that should not be public
        # This function changes the formatting to the following:
        # [engine]://[db_name]/[key]

        full_url = str(self.engine.url)
        engine_name = full_url.split("://")[0]
        db_name = full_url.split("/")[-1]
        return f"{engine_name}://{db_name}/{key[0]!s}"

    def _has_key(self, key):  # type: ignore[explicit-override] # FIXME
        sel = (
            sa.select(sa.func.count(sa.column("value")))
            .select_from(self._table)
            .where(
                sa.and_(
                    *(
                        getattr(self._table.columns, key_col) == val
                        for key_col, val in zip(self.key_columns, key)
                    )
                )
            )
        )
        try:
            with self.engine.begin() as connection:
                return connection.execute(sel).fetchone()[0] == 1
        except (IndexError, SQLAlchemyError) as e:
            logger.debug(f"Error checking for value: {e!s}")
            return False

    def list_keys(self, prefix=()):  # type: ignore[explicit-override] # FIXME
        columns = [sa.column(col) for col in self.key_columns]
        sel = (
            sa.select(*columns)
            .select_from(self._table)
            .where(
                sa.and_(
                    True,
                    *(
                        getattr(self._table.columns, key_col) == val
                        for key_col, val in zip(self.key_columns[: len(prefix)], prefix)
                    ),
                )
            )
        )
        with self.engine.begin() as connection:
            row_list: list[sqlalchemy.Row] = connection.execute(sel).fetchall()
        return [tuple(row) for row in row_list]

    def remove_key(self, key):  # type: ignore[explicit-override] # FIXME
        delete_statement = self._table.delete().where(
            sa.and_(
                *(
                    getattr(self._table.columns, key_col) == val
                    for key_col, val in zip(self.key_columns, key)
                )
            )
        )
        try:
            with self.engine.begin() as connection:
                return connection.execute(delete_statement)
        except SQLAlchemyError as e:
            raise gx_exceptions.StoreBackendError(  # noqa: TRY003 # FIXME CoP
                f"Unable to delete key: got sqlalchemy error {e!s}"
            )

    @property
    @override
    def config(self) -> dict:
        return self._config



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, cast

from great_expectations.compatibility import pyspark, sqlalchemy
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.exceptions import GreatExpectationsError
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.expectations.metrics.table_metric_provider import (
    TableMetricProvider,
)
from great_expectations.expectations.metrics.util import get_sqlalchemy_column_metadata

if TYPE_CHECKING:
    from great_expectations.execution_engine.sqlalchemy_batch_data import SqlAlchemyBatchData


class ColumnTypes(TableMetricProvider):
    metric_name = "table.column_types"
    value_keys = ("include_nested",)
    default_kwarg_values = {"include_nested": True}

    @metric_value(engine=PandasExecutionEngine)
    def _pandas(
        cls,
        execution_engine: PandasExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ):
        df, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )
        return [{"name": name, "type": dtype} for (name, dtype) in zip(df.columns, df.dtypes)]

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(
        cls,
        execution_engine: SqlAlchemyExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ):
        batch_id: Optional[str] = metric_domain_kwargs.get("batch_id")
        if batch_id is None:
            if execution_engine.batch_manager.active_batch_data_id is not None:
                batch_id = execution_engine.batch_manager.active_batch_data_id
            else:
                raise GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                    "batch_id could not be determined from domain kwargs and no active_batch_data is loaded into the "  # noqa: E501 # FIXME CoP
                    "execution engine"
                )

        batch_data: SqlAlchemyBatchData = cast(
            "SqlAlchemyBatchData",
            execution_engine.batch_manager.batch_data_cache.get(batch_id),
        )
        if batch_data is None:
            raise GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                "the requested batch is not available; please load the batch into the execution engine."  # noqa: E501 # FIXME CoP
            )

        return _get_sqlalchemy_column_metadata(execution_engine, batch_data)

    @metric_value(engine=SparkDFExecutionEngine)
    def _spark(
        cls,
        execution_engine: SparkDFExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ):
        df, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )
        spark_column_metadata = _get_spark_column_metadata(
            df.schema, include_nested=metric_value_kwargs["include_nested"]
        )
        return spark_column_metadata


def _get_sqlalchemy_column_metadata(
    execution_engine: SqlAlchemyExecutionEngine, batch_data: SqlAlchemyBatchData
):
    table_selectable: str | sqlalchemy.TextClause

    if sqlalchemy.Table and isinstance(batch_data.selectable, sqlalchemy.Table):  # type: ignore[truthy-function] # FIXME CoP
        table_selectable = batch_data.source_table_name or batch_data.selectable.name
        schema_name = batch_data.source_schema_name or batch_data.selectable.schema

    # if custom query was passed in
    elif sqlalchemy.TextClause and isinstance(batch_data.selectable, sqlalchemy.TextClause):  # type: ignore[truthy-function] # FIXME CoP
        table_selectable = batch_data.selectable
        schema_name = None
    else:
        table_selectable = batch_data.source_table_name or batch_data.selectable.name
        schema_name = batch_data.source_schema_name or batch_data.selectable.schema

    return get_sqlalchemy_column_metadata(
        execution_engine=execution_engine,
        table_selectable=table_selectable,  # type: ignore[arg-type] # FIXME CoP
        schema_name=schema_name,
    )


def _get_spark_column_metadata(field, parent_name="", include_nested=True):  # noqa: C901 #  too complex
    cols = []
    if parent_name != "":
        parent_name = f"{parent_name}."

    if pyspark.types and isinstance(field, pyspark.types.StructType):
        for child in field.fields:
            cols += _get_spark_column_metadata(
                child, parent_name=parent_name, include_nested=include_nested
            )
    elif pyspark.types and isinstance(field, pyspark.types.StructField):
        if include_nested and "." in field.name:
            # Only add backticks to escape dotted fields if they don't already exist
            if field.name.startswith("`") and field.name.endswith("`"):
                name = f"{parent_name}{field.name}"
            else:
                name = f"{parent_name}`{field.name}`"
        else:
            name = parent_name + field.name

        field_metadata = {"name": name, "type": field.dataType}
        cols.append(field_metadata)

        if (
            include_nested
            and pyspark.types
            and isinstance(field.dataType, pyspark.types.StructType)
        ):
            for child in field.dataType.fields:
                cols += _get_spark_column_metadata(
                    child,
                    parent_name=parent_name + field.name,
                    include_nested=include_nested,
                )
    else:
        raise ValueError("unrecognized field type")  # noqa: TRY003 # FIXME CoP

    return cols



import pandas

from great_expectations.metrics.column.distinct_values import ColumnDistinctValues
from great_expectations.metrics.metric_results import MetricErrorResult
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.metrics.conftest import ALL_DATA_SOURCES

DATA_FRAME = pandas.DataFrame(
    {
        "id": [1, 2, 3, 4],
        "number": [1, 2, 3, 4],
    },
)


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_DATA_SOURCES,
    data=DATA_FRAME,
)
def test_error_result(batch_for_datasource) -> None:
    batch = batch_for_datasource
    metric = ColumnDistinctValues(column="non_existent_column")
    metric_result = batch.compute_metrics(metric)
    assert isinstance(metric_result, MetricErrorResult)
    assert metric_result.value.exception_message is not None
    assert metric_result.value.exception_traceback is not None
